@@ -111,13 +111,17 @@ void PropSetFile::SetLine(const char *keyVal, bool unescape) {
 void PropSetFile::Unset(std::string_view key) {
 	if (key.empty())	// Empty keys are not supported
 		return;
-	mapss::iterator keyPos = props.find(std::string(key));
+	const mapss::iterator keyPos = props.find(key);
 	if (keyPos != props.end())
 		props.erase(keyPos);
 }
 
-bool PropSetFile::Exists(const char *key) const {
-	mapss::const_iterator keyPos = props.find(std::string(key));
+void PropSetFile::Clear() noexcept {
+	props.clear();
+}
+
+bool PropSetFile::Exists(std::string_view key) const {
+	const mapss::const_iterator keyPos = props.find(key);
 	if (keyPos != props.end()) {
 		return true;
 	} else {
@@ -130,11 +134,10 @@ bool PropSetFile::Exists(const char *key) const {
 	}
 }
 
-std::string PropSetFile::GetString(const char *key) const {
-	const std::string sKey(key);
+std::string_view PropSetFile::Get(std::string_view key) const {
 	const PropSetFile *psf = this;
 	while (psf) {
-		mapss::const_iterator keyPos = psf->props.find(sKey);
+		const mapss::const_iterator keyPos = psf->props.find(key);
 		if (keyPos != psf->props.end()) {
 			return keyPos->second;
 		}
@@ -144,6 +147,10 @@ std::string PropSetFile::GetString(const char *key) const {
 	return "";
 }
 
+std::string PropSetFile::GetString(std::string_view key) const {
+	return std::string(Get(key));
+}
+
 static std::string ShellEscape(const char *toEscape) {
 	std::string str(toEscape);
 	for (ptrdiff_t i = str.length()-1; i >= 0; --i) {
@@ -179,14 +186,14 @@ static std::string ShellEscape(const char *toEscape) {
 	return str;
 }
 
-std::string PropSetFile::Evaluate(const char *key) const {
-	if (strchr(key, ' ')) {
-		if (isprefix(key, "escape ")) {
-			std::string val = GetString(key+7);
+std::string PropSetFile::Evaluate(std::string_view key) const {
+	if (key.find(' ') != std::string_view::npos) {
+		if (StartsWith(key, "escape ")) {
+			std::string val(Get(key.substr(7)));
 			std::string escaped = ShellEscape(val.c_str());
 			return escaped;
-		} else if (isprefix(key, "= ")) {
-			const std::string sExpressions(key + 2);
+		} else if (StartsWith(key, "= ")) {
+			const std::string sExpressions(key.substr(2));
 			std::vector<std::string> parts = StringSplit(sExpressions, ';');
 			if (parts.size() > 1) {
 				bool equal = true;
@@ -197,15 +204,15 @@ std::string PropSetFile::Evaluate(const char *key) const {
 				}
 				return equal ? "1" : "0";
 			}
-		} else if (isprefix(key, "star ")) {
-			const std::string sKeybase(key + 5);
+		} else if (StartsWith(key, "star ")) {
+			const std::string sKeybase(key.substr(5));
 			// Create set of variables with values
 			mapss values;
 			// For this property set and all base sets
 			for (const PropSetFile *psf = this; psf; psf = psf->superPS) {
 				mapss::const_iterator it = psf->props.lower_bound(sKeybase);
 				while ((it != psf->props.end()) && (it->first.find(sKeybase) == 0)) {
-					mapss::iterator itDestination = values.find(it->first);
+					const mapss::iterator itDestination = values.find(it->first);
 					if (itDestination == values.end()) {
 						// Not present so add
 						values[it->first] = it->second;
@@ -219,18 +226,18 @@ std::string PropSetFile::Evaluate(const char *key) const {
 				combination += itV->second;
 			}
 			return combination;
-		} else if (isprefix(key, "scale ")) {
+		} else if (StartsWith(key, "scale ")) {
 			const int scaleFactor = GetInt("ScaleFactor", 100);
-			const char *val = key + 6;
+			std::string val(key.substr(6));
 			if (scaleFactor == 100) {
 				return val;
 			} else {
-				const int value = atoi(val);
+				const int value = IntegerFromString(val, 1);
 				return StdStringFromInteger(value * scaleFactor / 100);
 			}
 		}
 	} else {
-		return GetString(key);
+		return std::string(Get(key));
 	}
 	return "";
 }
@@ -241,15 +248,16 @@ std::string PropSetFile::Evaluate(const char *key) const {
 // for that, through a recursive function and a simple chain of pointers.
 
 struct VarChain {
-	VarChain(const char *var_=nullptr, const VarChain *link_=nullptr) noexcept : var(var_), link(link_) {}
+	VarChain() = default;
+	explicit VarChain(std::string_view var_, const VarChain *link_=nullptr) noexcept : var(var_), link(link_) {}
 
-	bool contains(const char *testVar) const {
-		return (var && (0 == strcmp(var, testVar)))
+	bool contains(std::string_view testVar) const {
+		return (var == testVar)
 		       || (link && link->contains(testVar));
 	}
 
-	const char *var;
-	const VarChain *link;
+	const std::string_view var;
+	const VarChain *link=nullptr;
 };
 
 static int ExpandAllInPlace(const PropSetFile &props, std::string &withVars, int maxExpands, const VarChain &blankVars = VarChain()) {
@@ -269,14 +277,14 @@ static int ExpandAllInPlace(const PropSetFile &props, std::string &withVars, int
 		}
 
 		std::string var(withVars, varStart + 2, varEnd - (varStart + 2));
-		std::string val = props.Evaluate(var.c_str());
+		std::string val = props.Evaluate(var);
 
-		if (blankVars.contains(var.c_str())) {
+		if (blankVars.contains(var)) {
 			val.clear(); // treat blankVar as an empty string (e.g. to block self-reference)
 		}
 
 		if (--maxExpands >= 0) {
-			maxExpands = ExpandAllInPlace(props, val, maxExpands, VarChain(var.c_str(), &blankVars));
+			maxExpands = ExpandAllInPlace(props, val, maxExpands, VarChain(var, &blankVars));
 		}
 
 		withVars.erase(varStart, varEnd-varStart+1);
@@ -288,34 +296,30 @@ static int ExpandAllInPlace(const PropSetFile &props, std::string &withVars, int
 	return maxExpands;
 }
 
-std::string PropSetFile::GetExpandedString(const char *key) const {
-	std::string val = GetString(key);
+std::string PropSetFile::GetExpandedString(std::string_view key) const {
+	std::string val(Get(key));
 	ExpandAllInPlace(*this, val, 200, VarChain(key));
 	return val;
 }
 
-std::string PropSetFile::Expand(const std::string &withVars, int maxExpands) const {
-	std::string val = withVars;
+std::string PropSetFile::Expand(std::string_view withVars, int maxExpands) const {
+	std::string val(withVars);
 	ExpandAllInPlace(*this, val, maxExpands);
 	return val;
 }
 
-int PropSetFile::GetInt(const char *key, int defaultValue) const {
+int PropSetFile::GetInt(std::string_view key, int defaultValue) const {
 	return IntegerFromString(GetExpandedString(key), defaultValue);
 }
 
-intptr_t PropSetFile::GetInteger(const char *key, intptr_t defaultValue) const {
+intptr_t PropSetFile::GetInteger(std::string_view key, intptr_t defaultValue) const {
 	return IntPtrFromString(GetExpandedString(key), defaultValue);
 }
 
-long long PropSetFile::GetLongLong(const char *key, long long defaultValue) const {
+long long PropSetFile::GetLongLong(std::string_view key, long long defaultValue) const {
 	return LongLongFromString(GetExpandedString(key), defaultValue);
 }
 
-void PropSetFile::Clear() noexcept {
-	props.clear();
-}
-
 /**
  * Get a line of input. If end of line escaped with '\\' then continue reading.
  */
@@ -403,12 +407,12 @@ PropSetFile::ReadLineState PropSetFile::ReadLine(const char *lineBuffer, ReadLin
 		} else if (value == "1") {
 			rls = ReadLineState::active;
 		} else {
-			rls = (GetInt(value.c_str()) != 0) ? ReadLineState::active : ReadLineState::conditionFalse;
+			rls = (GetInt(value) != 0) ? ReadLineState::active : ReadLineState::conditionFalse;
 		}
 	} else if (superPS && isprefix(lineBuffer, "match ")) {
 		// Don't match on stand-alone property sets like localiser
 		const std::string pattern = lineBuffer + strlen("match") + 1;
-		const std::string relPath = GetString("RelativePath");
+		const std::string relPath(Get("RelativePath"));
 		const bool matches = PathMatch(pattern, relPath);
 		rls = matches ? ReadLineState::active : ReadLineState::conditionFalse;
 	} else if (isprefix(lineBuffer, "import ")) {
@@ -477,13 +481,7 @@ bool StringEqual(std::string_view a, std::string_view b, bool caseSensitive) noe
 	if (caseSensitive) {
 		return a == b;
 	} else {
-		if (a.length() != b.length()) {
-			return false;
-		}
-		for (size_t i = 0; i < a.length(); i++) {
-			if (MakeUpperCase(a[i]) != MakeUpperCase(b[i]))
-				return false;
-		}
+		return EqualCaseInsensitive(a, b);
 	}
 	return true;
 }
@@ -524,45 +522,48 @@ bool MatchWild(std::string_view pattern, std::string_view text, bool caseSensiti
 	return false;
 }
 
-bool startswith(const std::string &s, const char *keybase) noexcept {
-	return isprefix(s.c_str(), keybase);
+bool MatchWildSet(std::string_view patternSet, std::string_view text, bool caseSensitive) {
+	while (!patternSet.empty()) {
+		const size_t sepPos = patternSet.find_first_of(';');
+		const std::string_view pattern = patternSet.substr(0, sepPos);
+		if (MatchWild(pattern, text, caseSensitive)) {
+			return true;
+		}
+		// Move to next
+		patternSet = (sepPos == std::string_view::npos) ? "" : patternSet.substr(sepPos + 1);
+	}
+	return false;
 }
 
 }
 
-std::string PropSetFile::GetWildUsingStart(const PropSetFile &psStart, const char *keybase, const char *filename) {
-	const std::string sKeybase(keybase);
+std::string_view PropSetFile::GetWildUsingStart(const PropSetFile &psStart, std::string_view keybase, std::string_view filename) {
 	const PropSetFile *psf = this;
 	while (psf) {
-		mapss::const_iterator it = psf->props.lower_bound(sKeybase);
-		while ((it != psf->props.end()) && startswith(it->first, sKeybase.c_str())) {
-			const std::string_view orgkeyfile = it->first.c_str() + sKeybase.length();
+		mapss::const_iterator it = psf->props.lower_bound(keybase);
+		while ((it != psf->props.end()) && StartsWith(it->first, keybase)) {
+			if (it->first == keybase) {
+				return it->second;
+			}
+			const std::string_view first = it->first;
+			const std::string_view orgkeyfile = first.substr(keybase.length());
 			std::string key;	// keyFile may point into key so key lifetime must cover keyFile
 			std::string_view keyFile = orgkeyfile;
 
-			if (orgkeyfile.find("$(") == 0) {
+			if (StartsWith(orgkeyfile, "$(")) {
 				// $(X) is a variable so extract X and find its value
 				const size_t endVar = orgkeyfile.find_first_of(')');
 				if (endVar != std::string_view::npos) {
-					const std::string var(orgkeyfile.substr(2, endVar-2));
-					key = psStart.GetExpandedString(var.c_str());
+					const std::string_view var = orgkeyfile.substr(2, endVar-2);
+					key = psStart.GetExpandedString(var);
 					keyFile = key;
 				}
 			}
 
-			while (!keyFile.empty()) {
-				const size_t sepPos = keyFile.find_first_of(';');
-				const std::string_view pattern = keyFile.substr(0, sepPos);
-				if (MatchWild(pattern, filename, caseSensitiveFilenames)) {
-					return it->second;
-				}
-				// Move to next
-				keyFile = (sepPos == std::string_view::npos) ? "" : keyFile.substr(sepPos + 1);
-			}
-
-			if (it->first == sKeybase) {
+			if (MatchWildSet(keyFile, filename, caseSensitiveFilenames)) {
 				return it->second;
 			}
+
 			++it;
 		}
 		// Failed here, so try in base property set
@@ -571,15 +572,15 @@ std::string PropSetFile::GetWildUsingStart(const PropSetFile &psStart, const cha
 	return "";
 }
 
-std::string PropSetFile::GetWild(const char *keybase, const char *filename) {
+std::string_view PropSetFile::GetWild(std::string_view keybase, std::string_view filename) {
 	return GetWildUsingStart(*this, keybase, filename);
 }
 
 // GetNewExpandString does not use Expand as it has to use GetWild with the filename for each
 // variable reference found.
 
-std::string PropSetFile::GetNewExpandString(const char *keybase, const char *filename) {
-	std::string withVars = GetWild(keybase, filename);
+std::string PropSetFile::GetNewExpandString(std::string_view keybase, std::string_view filename) {
+	std::string withVars(GetWild(keybase, filename));
 	size_t varStart = withVars.find("$(");
 	int maxExpands = 1000;	// Avoid infinite expansion of recursive definitions
 	while ((varStart != std::string::npos) && (maxExpands > 0)) {
@@ -587,10 +588,11 @@ std::string PropSetFile::GetNewExpandString(const char *keybase, const char *fil
 		if (varEnd == std::string::npos) {
 			break;
 		}
-		std::string var(withVars, varStart + 2, varEnd - varStart - 2);	// Subtract the $(
-		std::string val = GetWild(var.c_str(), filename);
-		if (var == keybase)
-			val.clear(); // Self-references evaluate to empty string
+		const std::string_view var = std::string_view(withVars).substr(
+			varStart + 2, varEnd - varStart - 2);	// Subtract the $(
+		std::string val;
+		if (var != keybase)	// Self-references evaluate to empty string
+			val = GetWild(var, filename);
 		withVars.replace(varStart, varEnd - varStart + 1, val);
 		varStart = withVars.find("$(");
 		maxExpands--;
@@ -602,7 +604,7 @@ std::string PropSetFile::GetNewExpandString(const char *keybase, const char *fil
  * Initiate enumeration.
  */
 bool PropSetFile::GetFirst(const char *&key, const char *&val) const {
-	mapss::const_iterator it = props.begin();
+	const mapss::const_iterator it = props.begin();
 	if (it != props.end()) {
 		key = it->first.c_str();
 		val = it->second.c_str();
